/*
  atv_vfd.c

  userspace HAL program to control a Altivar  VFD

  Copyright (C) 2007, 2008 Stephen Wille Padnos, Thoth Systems, Inc.

  Based on a work (test-modbus program, part of libmodbus) which is
  Copyright (C) 2001-2005 Stéphane Raimbault <stephane.raimbault@free.fr>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation, version 2.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.


  This is a userspace HAL program, which may be loaded using the halcmd "loadusr" command:
  loadusr atv_vfd
  There are several command-line options.  Options that have a set list of possible values may
  be set by using any number of characters that are unique.  For example, --rate 5 will use
  a baud rate of 57600, since no other available baud rates start with "5"
  -b or --bits <n> (default 8)
  Set number of data bits to <n>, where n must be from 5 to 8 inclusive
  -d or --device <path> (default /dev/ttyS0)
  Set the name of the serial device node to use
  -g or --debug
  Turn on debugging messages.  This will also set the verbose flag.  Debug mode will cause
  all modbus messages to be printed in hex on the terminal.
  -n or --name <string> (default atv_vfd)
  Set the name of the HAL module.  The HAL comp name will be set to <string>, and all pin
  and parameter names will begin with <string>.
  -p or --parity {even,odd,none} (defalt odd)
  Set serial parity to even, odd, or none.
  -r or --rate <n> (default 38400)
  Set baud rate to <n>.  It is an error if the rate is not one of the following:
  110, 300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200
  -s or --stopbits {1,2} (default 1)
  Set serial stop bits to 1 or 2
  -t or --target <n> (default 1)
  Set MODBUS target (slave) number.  This must match the device number you set on the VF-S11.
  -v or --verbose
  Turn on debug messages.  Note that if there are serial errors, this may become annoying.
  At the moment, it doesn't make much difference most of the time.


  $Id: atv_vfd.c,v 1.0 2010/04/10 13:56:11 mah Exp $
*/


#ifndef ULAPI
#error This is intended as a userspace component only.
#endif


#define DEBUG
//#undef DEBUG

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <getopt.h>
#include <math.h>
#include <signal.h>
#include "rtapi.h"
#include "hal.h"
#include "modbus.h"

// ATV registers
#define REG_CMD				0x2135  // Control Word CMD 8501 frequency
#define REG_ETA				0x0c81	// Status Word ETA 3201 frequency
#define REG_LFR				0x2136  // Frequency set point
#define REG_RFR				0x0c82	// Output freq
#define REG_RFRD			0x219c	// Output velocity RPM
#define REG_ULN				0x0c87	// Main voltage 0.1V
#define REG_LCR				0x0c84	// MOtor current 0.1A
#define REG_UOP				0x0c88	// Motor Voltage 1v
#define REG_OPR				0x0c8B	// Motor power 1%
#define REG_HMIS			0x0ca8  // Drive state 


#define ACTIVE_REF_CHANNEL		0x20F9  // Active reference channel bit 3 HMI bit 4 Modbus
#define ACTIVE_COM_CHANNEL		0x20FA  // Active command channel bit 3 HMI bit 4 Modbus
#define MAX_FREQ			0x0c20  // HSP Hi speed *0.1 0-1600Hz
#define MIN_FREQ			0x0C21  // LSP    Low speed *0.1 0-1600Hz
#define STD_MOT_FREQ			0x0BC7	// Standard motor frequency
#define RAT_MOT_VOLT			0x2581	// Rated motor voltage
#define RAT_MOT_CURR			0x2583 	// Rated motor current
#define RAT_MOT_REV			0x2584	// Rated motor speed RPM
#define STATUS				0x219B	// Status


//ETA 
#define STOP_KEY			0x4000  // Stop via STOP key from VFD
#define AT_SPEED			0x0400	// At speed
#define READY_TO_SWITCH_ON		0x0021	// Ready to switch on
#define SWITCH_ON			0x0023	// Switch on
#define SWITCH_ON_DISABLED		0x0040	// Switch on disabled (MUST DO SHUTDOWN)
#define OPERATION_ENABLED		0x0027  // operation enabled
#define QUICK_STOP_ACTIVE		0x0007  // Quick stop active
#define ABSENT_POWER			0x0010  // no power supply

//CMD
#define QUICK_STOP			0x0002	//quick stop
#define DISABLE_VOLTAGE			0x0000	// Disable Voltage 
#define SHUTDOWN			0x0006	//Shutdown
#define ENABLE_OPERATION		0x000f	// Enable operation


#define MODBUS_MIN_OK	10


int modbus_ok;						// set modbus-ok bit if last MODBUS_OK transactions went well
int debug, verbose,modbus_debug;
int slave = 1;

/* HAL data struct */
typedef struct {
    // status words from the VFDthe name of the HAL module. The HAL comp name will be set to <string>, and all pin and parameter
    // names will begin with <string>.  Maybe split these out sometime
    hal_bit_t 	*enable;				// enable vfd after load EMC and AXIS is ready
    hal_float_t     *max_freq;                              // max frequency
    hal_float_t	*min_freq;				// min frequency	
    hal_bit_t	*std_mot_freq;				// Standart motor frequency 0/1 50Hz/60Hz
    hal_float_t	*rat_mot_volt;				// Rated motor voltage
    hal_float_t	*rat_mot_curr;				// Rated motor current
    hal_float_t     *rat_mot_rev;				// Rated motor velocity RPM 
    hal_float_t     *out_freq;				// Output vfd frequency
    hal_float_t	*out_rev;				// Output velocity rpm
    hal_float_t	*main_volt;				// Main voltage 1.1V
    hal_float_t 	*motor_curr;				// Motor current 0.1A
    hal_float_t	*motor_volt;				// Motor voltage 1V
    hal_float_t 	*motor_power;				// Motor power 1%			
    hal_s32_t 	*status;				// ETA
    hal_bit_t	*modbus_ok;				// the last MODBUS_OK transactions returned successfully
    hal_bit_t	*spindle_on;				// spindle 1=on, 0=off
    hal_bit_t	*is_estop;				// Emergency stop from EMC 1=active
    hal_bit_t	*estop;					// Emergencz stop from VFD 1=active
    hal_bit_t	mdb_ready;				// Modbus ready (cmd.channel=modbus ref.channel=modbus)
    hal_float_t	*speed_command;				// speed input
    hal_bit_t       *at_speed;	            		// when drive speed_command == out_freq and running 
    hal_s32_t 	retval;
    hal_float_t	looptime;
} haldata_t;

static int done;
char *modname = "atv_vfd";

static struct option long_options[] = {
    {"bits", 1, 0, 'b'},
    {"device", 1, 0, 'd'},
    {"debug", 0, 0, 'g'},
    {"help", 0, 0, 'h'},
    {"name", 1, 0, 'n'},
    {"parity", 1, 0, 'p'},
    {"rate", 1, 0, 'r'},
    {"stopbits", 1, 0, 's'},
    {"target", 1, 0, 't'},
    {"verbose", 0, 0, 'v'},
    {"modbus-debug", 0, 0, 'M'},
    {0,0,0,0}
};

static char *option_string = "b:d:hn:p:r:s:t:vgPM";

static char *bitstrings[] = {"5", "6", "7", "8", NULL};
static char *paritystrings[] = {"even", "odd", "none", NULL};
static char *ratestrings[] = {"110", "300", "600", "1200", "2400", "4800", "9600",
			      "19200", "38400", "57600", "115200", NULL};
static char *stopstrings[] = {"1", "2", NULL};

static void quit(int sig) {
    done = 1;
}

int match_string(char *string, char **matches) {
    int len, which, match;
    which=0;
    match=-1;
    if ((matches==NULL) || (string==NULL)) return -1;
    len = strlen(string);
    while (matches[which] != NULL) {
	if ((!strncmp(string, matches[which], len)) && (len <= strlen(matches[which]))) {
	    if (match>=0) return -1;        // multiple matches
	    match=which;
	}
	++which;
    }
    return match;
}

int write_data(modbus_param_t *param, int slave, haldata_t *haldata)
{
    int retval = 0;
    int cmd_reg;				//Command register for CMD/
    hal_float_t hzcalc;

    hzcalc = 500/(*haldata->rat_mot_rev)*(*haldata->speed_command);

    if (*(haldata->is_estop)){
	cmd_reg = QUICK_STOP ;
    }
    else	{
	cmd_reg = SHUTDOWN ;	
	if  (*(haldata->spindle_on)){
	    if (haldata->mdb_ready)	{	
		if ((*(haldata->status) & OPERATION_ENABLED ) == OPERATION_ENABLED){
		    cmd_reg = ENABLE_OPERATION;
		}
		else {
		    if ((*(haldata->status) & READY_TO_SWITCH_ON)  ==  READY_TO_SWITCH_ON){
			cmd_reg = ENABLE_OPERATION;
		    }
		}
	    }
	}
	else { 
	    cmd_reg = DISABLE_VOLTAGE ;	
	}
		
    }
    preset_single_register(param, slave, REG_LFR,hzcalc);
    preset_single_register(param, slave, REG_CMD,cmd_reg) ;
    return retval;
}

void usage(int argc, char **argv) {
    printf("Usage:  %s [options]\n", argv[0]);
    printf(
	   "This is a userspace HAL program, typically loaded using the halcmd \"loadusr\" command:\n"
	   "    loadusr atv_vfd\n"
	   "There are several command-line options.  Options that have a set list of possible values may\n"
	   "    be set by using any number of characters that are unique.  For example, --rate 5 will use\n"
	   "    a baud rate of 57600, since no other available baud rates start with \"5\"\n"
	   "-b or --bits <n> (default 8)\n"
	   "    Set number of data bits to <n>, where n must be from 5 to 8 inclusive\n"
	   "-d or --device <path> (default /dev/ttyS0)\n"
	   "    Set the name of the serial device node to use\n"
	   "-g or --debug\n"
	   "    Turn on debugging messages.  This will also set the verbose flag.  Debug mode will cause\n"
	   "    all modbus messages to be printed in hex on the terminal.\n"
	   "-M or --debug\n"
	   "    Turn on debugging messages.  This will also set the verbose flag.  Debug mode will cause\n"
	   "    all modbus messages to be printed in hex on the terminal.\n"
	   "-n or --name <string> (default atv_vfd)\n"
	   "    Set the name of the HAL module.  The HAL comp name will be set to <string>, and all pin\n"
	   "    and parameter names will begin with <string>.\n"
	   "-p or --parity {even,odd,none} (defalt even)\n"
	   "    Set serial parity to even, odd, or none.\n"
	   "-r or --rate <n> (default 19200)\n"
	   "    Set baud rate to <n>.  It is an error if the rate is not one of the following:\n"
	   "    110, 300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200\n"
	   "-s or --stopbits {1,2} (default 1)\n"
	   "    Set serial stop bits to 1 or 2\n"
	   "-t or --target <n> (default 1)\n"
	   "    Set MODBUS target (slave) number.  This must match the device number you set on the VF-S11.\n"
	   "-v or --verbose\n"
	   "    Turn on debug messages.  Note that if there are serial errors, this may become annoying.\n"
	   "    At the moment, it doesn't make much difference most of the time.\n");
}

int read_data(modbus_param_t *param, int slave, haldata_t *hal_data_block)
{
    int retval, curr_reg, val, val2, status_reg, freq_reg, i;
    short int val3;
    static int once = 1;
    static int pollcount = 0;

    /* can't do anything with a null HAL data block */
    if (hal_data_block == NULL)
	return -1;
    /* but we can signal an error if the other params are null */
    if (param == NULL) {
	return -1;
    }

    // some data are read only once at startup
    if (once) {
#ifdef DEBUG
	if (debug)
	    printf("read_data: reading setup parameters:\n");
#endif
	curr_reg = MIN_FREQ;
        if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	    goto failed;
        *(hal_data_block->min_freq) = val * 0.1;

        curr_reg = MAX_FREQ;
        if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	    goto failed;
        *(hal_data_block->max_freq) = val*0.1 ;

        curr_reg = STD_MOT_FREQ;
        if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	    goto failed;
        *(hal_data_block->std_mot_freq) = val ;
	curr_reg = RAT_MOT_VOLT;
        if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	    goto failed;
        *(hal_data_block->rat_mot_volt) = val ;
        
	curr_reg = RAT_MOT_CURR;
        if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	    goto failed;
        *(hal_data_block->rat_mot_curr) = val * 0.1;

	curr_reg = RAT_MOT_REV;
        if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	    goto failed;
        *(hal_data_block->rat_mot_rev) = val ;

#ifdef DEBUG
	if (debug)
	    printf("read_data: read setup parameters - OK.\n");
#endif
	once = 0;
    }

    curr_reg = REG_RFR;
    if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	goto failed;
    val3 = val;
    *(hal_data_block->out_freq) = val3*0.1  ;


    curr_reg = REG_RFRD;
    if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	goto failed;
    val3 = val;
    *(hal_data_block->out_rev) = abs(val3)  ;

    curr_reg = REG_ETA;
    if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	goto failed;
    *(hal_data_block->status) = val ;
    
    if (*(hal_data_block->status) & STOP_KEY) {
	*(hal_data_block->estop) = 1;
    } else *(hal_data_block->estop) = 0 ;

    if (val & AT_SPEED) {
	*(hal_data_block->at_speed) = 1;	
    } else *(hal_data_block->at_speed) = 0;

    //	Set up mdb_ready
    curr_reg = ACTIVE_REF_CHANNEL;
    if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	goto failed;
    curr_reg = ACTIVE_COM_CHANNEL;
    if ((retval = read_holding_registers(param, slave, curr_reg,1, &val2)) != 1)
	goto failed;
    if ((val & 0x0008) && (val2 & 0x0008))  {
	hal_data_block->mdb_ready = 1;
    }  else {
	hal_data_block->mdb_ready = 0;
    }

    curr_reg = REG_LCR;
    if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	goto failed;
    *(hal_data_block->motor_curr) = val*0.1;

    curr_reg = REG_UOP ;
    if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	goto failed;
    *(hal_data_block->motor_volt) = val;

    curr_reg = REG_OPR;
    if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	goto failed;
    *(hal_data_block->motor_power) = val;

    curr_reg = REG_ULN;
    if ((retval = read_holding_registers(param, slave, curr_reg,1, &val)) != 1)
	goto failed;
    *(hal_data_block->main_volt) = val*0.1;

#ifdef DEBUG
    if (debug)
	printf("read_data: status_reg=%4.4x freq_reg=%4.4x\n", status_reg, freq_reg);
#endif
 failed:
    hal_data_block->retval = retval;
    return retval;
}

int main(int argc, char **argv)
{
    int retval;
    modbus_param_t mb_param;
    haldata_t *haldata;
    int hal_comp_id;
    struct timespec loop_timespec, remaining;
    int baud, bits, stopbits;
    char *device, *parity, *endarg;
    int opt;
    int argindex, argvalue;
    done = 0;

    // assume that nothing is specified on the command line
    baud = 19200;
    bits = 8;
    stopbits = 1;
    debug = 0;
    verbose = 0;
    device = "/dev/ttyS0";
    parity = "even";
    slave = 1;

    // process command line options
    while ((opt=getopt_long(argc, argv, option_string, long_options, NULL)) != -1) {
	switch(opt) {
	case 'b':   // serial data bits, probably should be 8 (and defaults to 8)
	    argindex=match_string(optarg, bitstrings);
	    if (argindex<0) {
		printf("atv_vfd: ERROR: invalid number of bits: %s\n", optarg);
		retval = -1;
		goto out_noclose;
	    }
	    bits = atoi(bitstrings[argindex]);
	    break;
	case 'd':   // device name, default /dev/ttyS0
	    // could check the device name here, but we'll leave it to the library open
	    if (strlen(optarg) > FILENAME_MAX) {
		printf("at_vfd: ERROR: device node name is too long: %s\n", optarg);
		retval = -1;
		goto out_noclose;
	    }
	    device = strdup(optarg);
	    break;
	case 'g':
	    debug = 1;
	    verbose = 1;
	    break;
	case 'M':
	    modbus_debug = 1;
	    break;
	case 'n':   // module base name
	    if (strlen(optarg) > HAL_NAME_LEN-20) {
		printf("atv_vfd: ERROR: HAL module name too long: %s\n", optarg);
		retval = -1;
		goto out_noclose;
	    }
	    modname = strdup(optarg);
	    break;
	case 'p':   // parity, should be a string like "even", "odd", or "none"
	    argindex=match_string(optarg, paritystrings);
	    if (argindex<0) {
		printf("atv_vfd: ERROR: invalid parity: %s\n", optarg);
		retval = -1;
		goto out_noclose;
	    }
	    parity = paritystrings[argindex];
	    break;
	case 'r':   // Baud rate, 19200 default
	    argindex=match_string(optarg, ratestrings);
	    if (argindex<0) {
		printf("atv_vfd: ERROR: invalid baud rate: %s\n", optarg);
		retval = -1;
		goto out_noclose;
	    }
	    baud = atoi(ratestrings[argindex]);
	    break;
	case 's':   // stop bits, defaults to 1
	    argindex=match_string(optarg, stopstrings);
	    if (argindex<0) {
		printf("atv_vfd: ERROR: invalid number of stop bits: %s\n", optarg);
		retval = -1;
		goto out_noclose;
	    }
	    stopbits = atoi(stopstrings[argindex]);
	    break;
	case 't':   // target number (MODBUS ID), default 1
	    argvalue = strtol(optarg, &endarg, 10);
	    if ((*endarg != '\0') || (argvalue < 1) || (argvalue > 254)) {
		printf("atv_vfd: ERROR: invalid slave number: %s\n", optarg);
		retval = -1;
		goto out_noclose;
	    }
	    slave = argvalue;
	    break;

	case 'v':   // verbose mode (print modbus errors and other information), default 0
	    verbose = 1;
	    break;
	case 'h':
	default:
	    usage(argc, argv);
	exit(0);
	break;
	}
    }

    printf("%s: device='%s', baud=%d, bits=%d, parity='%s', stopbits=%d, address=%d, verbose=%d\n",
	   modname, device, baud, bits, parity, stopbits, slave, debug);

    /* point TERM and INT signals at our quit function */
    /* if a signal is received between here and the main loop, it should prevent
       some initialization from happening */
    signal(SIGINT, quit);
    signal(SIGTERM, quit);

    /* Assume 19.2k E-8-1 serial settings, device 1 */
    modbus_init_rtu(&mb_param, device, baud, parity, bits, stopbits, verbose);
    mb_param.debug = modbus_debug;

    /* the open has got to work, or we're out of business */
    if (((retval = modbus_connect(&mb_param))!=0) || done) {
	printf("%s: ERROR: couldn't open serial device\n", modname);
	goto out_noclose;
    }
    /* create HAL component */
    hal_comp_id = hal_init(modname);
    if ((hal_comp_id < 0) || done) {
	printf("%s: ERROR: hal_init failed\n", modname);
	retval = hal_comp_id;
	goto out_close;
    }

    /* grab some shmem to store the HAL data in */
    haldata = (haldata_t *)hal_malloc(sizeof(haldata_t));
    if ((haldata == 0) || done) {
	printf("%s: ERROR: unable to allocate shared memory\n", modname);
	retval = -1;
	goto out_close;
    }

    retval = hal_pin_bit_newf(HAL_IN, &(haldata->enable), hal_comp_id, "%s.enable", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_OUT, &(haldata->max_freq), hal_comp_id, "%s.max-freq", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_OUT, &(haldata->min_freq), hal_comp_id, "%s.min-freq", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_bit_newf(HAL_OUT, &(haldata->std_mot_freq), hal_comp_id, "%s.std-mot-freq", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_OUT, &(haldata->out_rev), hal_comp_id, "%s.out-rev", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_OUT, &(haldata->main_volt), hal_comp_id, "%s.main-volt", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_OUT, &(haldata->motor_curr), hal_comp_id, "%s.motor-curr", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_OUT, &(haldata->motor_volt), hal_comp_id, "%s.motor-volt", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_OUT, &(haldata->motor_power), hal_comp_id, "%s.motor-power", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_OUT, &(haldata->rat_mot_volt), hal_comp_id, "%s.rat-mot-volt", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_OUT, &(haldata->rat_mot_curr), hal_comp_id, "%s.rat-mot-curr", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_OUT, &(haldata->rat_mot_rev), hal_comp_id, "%s.rat-mot-rev", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_s32_newf(HAL_OUT, &(haldata->status), hal_comp_id, "%s.status", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_IN, &(haldata->speed_command), hal_comp_id, "%s.speed-command", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_float_newf(HAL_OUT, &(haldata->out_freq), hal_comp_id, "%s.out-freq", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_bit_newf(HAL_OUT, &(haldata->modbus_ok), hal_comp_id, "%s.modbus-ok", modname); 
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_bit_newf(HAL_OUT, &(haldata->estop), hal_comp_id, "%s.estop", modname); 
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_bit_newf(HAL_IN, &(haldata->is_estop), hal_comp_id, "%s.is-estop", modname); 
    if (retval!=0) goto out_closeHAL;
    retval = hal_param_float_newf(HAL_RW, &(haldata->looptime), hal_comp_id, "%s.loop-time", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_bit_newf(HAL_IN, &(haldata->spindle_on), hal_comp_id, "%s.spindle-on", modname);
    if (retval!=0) goto out_closeHAL;
    retval = hal_pin_bit_newf(HAL_OUT, &(haldata->at_speed), hal_comp_id, "%s.at-speed", modname);
    if (retval!=0) goto out_closeHAL;
    /* make default data match what we expect to use */
    *(haldata->max_freq) = 0;
    *(haldata->min_freq) = 0;
    *(haldata->std_mot_freq) = 0;
    *(haldata->out_rev) = 0;
    haldata->looptime = 0.1;
    hal_ready(hal_comp_id);
    *(haldata->modbus_ok) = 0;
    *(haldata->main_volt) = 0;
    *(haldata->spindle_on) = 0;
    *(haldata->at_speed) = 0;
    /* here's the meat of the program.  loop until done (which may be never) */

    // wait for enable signal after load axis
    while (!*(haldata->enable)) {	
	// do nothing
	sleep(1);  // be nice to CPU -mah
    }

    while (done == 0) {
	read_data(&mb_param, slave, haldata);
	modbus_ok = haldata->retval==1 ?  (modbus_ok + 1) : 0 ;
	if (modbus_ok > MODBUS_MIN_OK) {
	    *(haldata->modbus_ok) = 1;
	} else {
	    *(haldata->modbus_ok) = 0;
	}
#ifdef DEBUG
	if (debug && haldata->retval)
	    printf("main: read_data retval=%d\n", haldata->retval);
#endif
	write_data(&mb_param, slave, haldata);
#ifdef DEBUG
	if (debug && haldata->retval)
	    if (debug)
		printf("main: write_data retval=%d\n", haldata->retval);
#endif
	/* don't want to scan too fast, and shouldn't delay more than a few seconds */
	if (haldata->looptime < 0.001) haldata->looptime = 0.001;
	if (haldata->looptime > 2.0) haldata->looptime = 2.0;
	loop_timespec.tv_sec = (time_t)(haldata->looptime);
	loop_timespec.tv_nsec = (long)((haldata->looptime - loop_timespec.tv_sec) * 1000000000l);
	nanosleep(&loop_timespec, &remaining);
    }

    // clear the command register (control and frequency override) so panel operation gets reactivated
    preset_single_register(&mb_param, slave, 0x2136,0);
    if ((retval = preset_single_register(&mb_param, slave, 0x2135, 0)) != 1) {
	// not much we can do about it here, nice try
#ifdef DEBUG
	if (debug)
	    printf("main: failed to clear command register, retval=%d\n", retval);
#endif
    } else {
	if (debug)
	    printf("main: cleared command register - OK\n");
    }
    retval = 0;	/* if we get here, then everything is fine, so just clean up and exit */

 out_closeHAL:
    hal_exit(hal_comp_id);
 out_close:
    modbus_close(&mb_param);
 out_noclose:
    return retval;
}
